﻿var fs = require('fs');
var events = require('events');
var moment = require('moment');

//---------------------------------------------
// Initializing Logging infrastructure
//#region Logging Infrastructure
// Start the logger
var log4js = require('log4js');
var accesslogger = log4js.getLogger("accesscontrol");
//accesslogger.setLevel(log4js.levels.INFO);// Toggle to this if you want just basic log info
accesslogger.setLevel(log4js.levels.ALL);// Toggle to this if you want all log info
//#endregion Logging Infrastructure

//---------------------------------------------
// Access Control Definitions
//#region Access Control Defintions
var AccessControlUsers = [];// Holds the list of Access Control Users
var AccessControlPorts = [];// Holds the list of Device Ports
var AccessControlIo;// Holds socket io object
var AccessControlNetSNMP;// Holds NetSNMP object
var AccessControlDatabase;// Holds database

var InvalidEntries = 0;

var AccessControlKeypadStateEnum = {
    NormalState: 0,
    ValidEntry: 1,
    InvalidEntry: -1
}

var _this = this;// To allow access to exported functions
//#endregion Access Control Defintions

//#region Access Control Events

var AccessControlEvents = new events.EventEmitter;
exports.AccessControlEvents = AccessControlEvents;

AccessControlEvents.on("DeviceEvent", function (device, data) {
    console.log('accesscontrol.AccessControlEvents.on: Port=' + JSON.stringify(device) + ' Data=' + JSON.stringify(data));
    
    switch (device.DeviceType) {

        case "Access Keypad":
            var validcode = false;
            for (i = 0; i < AccessControlUsers.length; i++) {
                if (AccessControlUsers[i].Code == data.KeypadCode) {
                    console.log("accesscontrol.AccessControlEvents.on: Valid code for user=" + AccessControlUsers[i].Name);
                    console.log("accesscontrol.AccessControlEvents.on: AccessControlUsers[" + i + "].Enabled=" + AccessControlUsers[i].Enabled);
                    console.log("accesscontrol.AccessControlEvents.on: AccessControlUsers[" + i + "].Name=" + AccessControlUsers[i].Name);
                    console.log("accesscontrol.AccessControlEvents.on: AccessControlUsers[" + i + "].Entries=" + AccessControlUsers[i].Entries);
                    
                    // Only send this event if the user is enabled
                    if (AccessControlUsers[i].Enabled === "TRUE") {
                        var user = AccessControlUsers[i].Name;
                        
                        ActivateDoorStrike(user, device, data);
                        
                        // Update the User's Entries
                        AccessControlUsers[i].Entries++;//Increment the access entries for this user
                        var event = { "DeviceType": "System", "Name": user, "TimeChanged": data.TimeChanged, "Entries": AccessControlUsers[i].Entries };
                        AccessControlIo.sockets.emit("AccessControl", event);

                        // Update the database with this entry
                        console.log("accesscontrol.AccessControlEvents.on: Valid code for user=" + JSON.stringify(AccessControlUsers));
                        var dbcommand = "UPDATE AccessControlUsers SET Entries=" + AccessControlUsers[i].Entries + ", TimeChanged='" + data.TimeChanged + "' WHERE rowid='" + AccessControlUsers[i].Id + "'";
                        console.log("accesscontrol.AccessControlEvents.on: Update=" + dbcommand);
                        AccessControlDatabase.run(dbcommand, function (err) {
                            if (err !== null) {
                                throw logger.error('accesscontrol.AccessControlEvents.on: User=' + AccessControlUsers[i].Name + ' - Error (Update)=' + JSON.stringify(err));
                            }                            
                        });

                    }
                    validcode = true;
                    break;
                }
            }
            if (validcode == false) {
                // Send the 'Access Keypad' event with name of *Invalid User* and State=InvalidEntry
                console.log("accesscontrol.ActivateDoorStrike: event=" + JSON.stringify(event));
                accesslogger.info("Invalid Code Entered: User=*Unknown*");
                var event = { "DeviceType": "Access Keypad", "Name": "* Unknown *", "TimeChanged": data.TimeChanged, "State": AccessControlKeypadStateEnum.InvalidEntry };
                SendSocketAndSnmp(device, event)

                // Update InvalidEntries                
                InvalidEntries++;// Increment the invalid entries
                var event = { "DeviceType": "System", "Name": "Invalid", "TimeChanged": data.TimeChanged, "Entries": InvalidEntries };
                AccessControlIo.sockets.emit("AccessControl", event);
                
                setTimeout(function () {
                    TimeChanged = moment();

                    // Send the Access Keypad event with name of System and State=NormalState
                    accesslogger.info("Normal State Entered: User=System");
                    var event = { "DeviceType": "Access Keypad", "Name": "System", "TimeChanged": TimeChanged, "State": AccessControlKeypadStateEnum.NormalState };
                    SendSocketAndSnmp(device, event)
                }, 2000);
            }
            break;
    
        default:
            console.log('accesscontrol.AccessControlEvents.on: Unhandled Device');
            break;
    }
});

function ActivateDoorStrike(user, device, data) {
    console.log("accesscontrol.ActivateDoorStrike: Device=" + JSON.stringify(device) + " data=" + JSON.stringify(data));
    
    // Find a door strike to enable    
    for (port = 0; port < AccessControlPorts.length; port++) {
        if (AccessControlPorts[port].DeviceType == "Door Strike") {
            // Send the 'Access Keypad' event with name of User and State=ValidEntry
            accesslogger.info("Valid Code Entered: User=" + user);
            var event = { "DeviceType": "Access Keypad", "Name": user, "TimeChanged": data.TimeChanged, "State": AccessControlKeypadStateEnum.ValidEntry };
            SendSocketAndSnmp(device, event)
            
            // Send the 'Door Strike' event with State=true
            var event = { "DeviceType": "Door Strike", "TimeChanged": data.TimeChanged, "State": true };
            AccessControlIo.sockets.emit("AccessControl", event);
            console.log("accesscontrol.ActivateDoorStrike: event=" + JSON.stringify(event));
            
            setTimeout(function () {
                TimeChanged = moment();
                
                // Send the Access Keypad event with name of System and State=NormalState
                accesslogger.info("Normal State Entered: User=System");
                var event = { "DeviceType": "Access Keypad", "Name": "System", "TimeChanged": TimeChanged, "State": AccessControlKeypadStateEnum.NormalState };
                SendSocketAndSnmp(device, event)
                
                // Send the 'Door Strike' event with State=false
                var event = { "DeviceType": "Door Strike", "TimeChanged": TimeChanged, "State": false };
                AccessControlIo.sockets.emit("AccessControl", event);
                console.log("accesscontrol.ActivateDoorStrike: event=" + JSON.stringify(event));
            }, 2000);
        }
    }
}
// Send updates to clients
function SendSocketAndSnmp(device, event) {
    console.log("accesscontrol.SendSocketAndSnmp: Device=" + JSON.stringify(device) + " event=" + JSON.stringify(event));
    
    // SOCKET.IO: Send changed value to all attached clients
    AccessControlIo.sockets.emit("AccessControl", event);
    
    // SNMP: Send SNMP Trap/Notification if threshold(s) crossed
    SendSNMP(event);
}

//#endregion Access Control Events

//---------------------------------------------
// SNMP infrastructure
//#region SNMP Infrastructure
// Initialize the snmpHostSessionsArray used for send SNMP messages
var SnmpHostSessionsArray = [];// Instance array for list of SNMP host sessions
exports.SetSnmpHostSessionsArray = function (snmpHostSessions, logger, callback) {
    
    SnmpHostSessionsArray = snmpHostSessions.slice(0);// Copy array
    logger.info("accesscontrol.SetSnmpHostSessionsArray: snmpHostSessions=" + snmpHostSessions.length + "  SnmpHostSessionsArray=" + SnmpHostSessionsArray.length);
    
    if (typeof callback == "function") {
        callback();
    }
}

function SendSNMP(event) {
    // SNMP definitions - also refer to the TASC-PEAK.MIB file for detailed information
    
    //console.log("accesscontrol.SendSNMP: Sending SNMP Trap for event=" + JSON.stringify(event));
    //console.log("accesscontrol.SendSNMP: AccessControlNetSnmp=" + JSON.stringify(AccessControlNetSNMP));
    
    //      AccessControlKeypadNotification NOTIFICATION-TYPE
    //      OBJECTS {
    //          PeakTimeStamp,
    //		    PeakHostName,
    //  	    AccessControlUser,
    //  	    AccessControlKeypadState
    //      }
    //      STATUS current
    //      DESCRIPTION "Access Control Keypad notification"
    //      -- 1.3.6.1.4.1.11902.4.100.10.1
    //      ::= { AccessControlNotifications 1 }
    var trapOid = "1.3.6.1.4.1.11902.4.100.10.1";
    
    var varbinds = [
        {
            //  PeakTimeStamp OBJECT-TYPE
            //      SYNTAX DateAndTime
            //      MAX-ACCESS not-accessible
            //      STATUS current
            //      DESCRIPTION "Date and Time at which the Site produced the event."
            //      -- 1.3.6.1.4.1.11902.4.1.1.1
            //      ::= { PeakGeneralOIDs 1 }
            
            oid: "1.3.6.1.4.1.11902.4.2.1.1",
            type: AccessControlNetSNMP.ObjectType.OctetString,
            value: new Buffer(moment().toString())
        },
        {
            //  PeakHostName OBJECT-TYPE
            //      SYNTAX OCTET STRING(SIZE(1..32))
            //      MAX-ACCESS read-only
            //      STATUS current
            //      DESCRIPTION "Peak Host Name"
            //      -- 1.3.6.1.4.1.11902.4.1.1.2
            //      ::= { PeakGeneralOIDs 2 }
            
            oid: "1.3.6.1.4.1.11902.4.2.1.2",
            type: AccessControlNetSNMP.ObjectType.OctetString,
            value: new Buffer("Peak")
        },
        {
            //  AccessControlUser OBJECT-TYPE
            //      SYNTAX OCTET STRING(SIZE(1..32))
            //      MAX-ACCESS not-accessible
            //      STATUS current
            //      DESCRIPTION "Access Control Keypad User"
            //      -- 1.3.6.1.4.1.11902.4.1.10.1
            //      ::= { PeakAccessControlOIDs 1 }
            oid: "1.3.6.1.4.1.11902.4.2.10.1",
            type: AccessControlNetSNMP.ObjectType.OctetString,
            value: new Buffer(event.Name)
        },
        {
            //  AccessControlKeypadState OBJECT-TYPE
            //      SYNTAX AccessControlKeypadStateEnum
            //      MAX-ACCESS not-accessible
            //      STATUS current
            //      DESCRIPTION "Access Control Keypad State"
            //      -- 1.3.6.1.4.1.11902.4.1.10.2
            //      ::= { PeakAccessControlOIDs 2 }
            oid: "1.3.6.1.4.1.11902.4.2.10.2",
            type: AccessControlNetSNMP.ObjectType.Integer,
            value: parseInt(event.State)
        }
    ];
    
    // Send the SNMP notification
    for (host = 0; host < SnmpHostSessionsArray.length; host++) {
        // Skip if this host is not enabled
        if (SnmpHostSessionsArray[host].Enabled === "FALSE") continue;
        
        // Send trap to host ...
        IPAddress = SnmpHostSessionsArray[host].Session.target;
        console.log("accesscontrol.SendSocketAndSnmp: IP Address=" + IPAddress);
        SnmpHostSessionsArray[host].Session.trap(trapOid, varbinds, function (error) {
            if (error)
                console.log("accesscontrol.SendSNMP: Send SNMP Error=" + error);
            else
                console.log("accesscontrol.SendSNMP: Sending to SNMP Host @ " + IPAddress + " - User=" + varbinds[2].value + " State=" + varbinds[3].value);
        });
    }
}
//#endregion SNMP Infrastructure

//#region Initialize Access Control Module
// Initialize 
exports.InitializeAccessControl = function (db, io, netsnmp, logger, callback) {
    
    // Set the io connection
    AccessControlIo = io;
    AccessControlNetSNMP = netsnmp;
    AccessControlDatabase = db;
    
    logger.info('AccessControl.InitializeAccessControl: netsnmp=' + JSON.stringify(netsnmp));
    
    // Try getting the Access Control Users table
    db.get("SELECT name FROM sqlite_master WHERE type='table' AND name='AccessControlUsers'", function (err, rows) {
        if (err !== null) {
            logger.error('AccessControl.InitializeAccessControl - Initialize Error (Select)=' + JSON.stringify(err));
        }
        else if (rows === undefined) {
            // We read nothing, so let's create the table ...
            logger.info("AccessControl.InitializeAccessControl - Initialize by creating database table ...");
            db.exec("CREATE TABLE if not exists AccessControlUsers (Name TEXT, Enabled TEXT, Entries INTEGER, Code TEXT, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    logger.error('AccessControl.InitializeAccessControl - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    if (typeof callback == "function") {
                        logger.info('AccessControl.InitializeAccessControl - Initialized new database table');
                        callback(AccessControlUsers);
                    }
                }
            });
        }
        else {
            // Read from the database and populate the AccessControlUsers array
            logger.info("AccessControl.InitializeAccessControl - Read Access Control Users from database ...");
            numberOfAccessControlUsers = 0;
            AccessControlUsers.length = 0;
            db.each("SELECT rowid AS Id, Name, Enabled, Entries, Code, TimeChanged FROM AccessControlUsers", function (err, row) {
                if (err !== null) {
                    logger.error('AccessControl.InitializeAccessControl - Initialize Error (Select)=' + JSON.stringify(err));
                } else {
                    // Copy each row into AccessControlUsers array
                    logger.debug('AccessControl.InitializeAccessControl - Processing Host[' + (numberOfAccessControlUsers + 1) + ']=' + row.Name);
                    AccessControlUsers.push({
                        Id : row.Id,
                        Name : row.Name,
                        Enabled : row.Enabled,
                        Entries : row.Entries,
                        Code : row.Code,
                        TimeChanged : row.TimeChanged
                    });
                    logger.info("AccessControl.InitializeAccessControl - Read AccessControlUsers from Database[" + numberOfAccessControlUsers + "]=" + JSON.stringify(AccessControlUsers[numberOfAccessControlUsers]));
                    numberOfAccessControlUsers++;
                }
            }, function () {
                // Retrurn the complete list of AccessControlUsers
                logger.info('AccessControl.InitializeAccessControl - Initialized from existing database table');
                if (typeof callback == "function") {
                    callback(AccessControlUsers);
                }
            });
        }
    });
}

exports.UpdateAccessControlPorts = function (ports) {
    AccessControlPorts = ports;
    console.log("accesscontrol.UpdateAccesControlPorts: ports=" + JSON.stringify(ports));
}

//#endregion Initialize Access Control Module

//#region Access Control Users

// Get Access Control Users
exports.GetAccessControlUsers = function (db, io, logger, callback) {
    // Get the complete list of AccessControlUsers
    AccessControlUsers = [];
    db.all("SELECT rowid as Id, Name, Enabled, Entries, Code, TimeChanged FROM AccessControlUsers ORDER BY Id", function (err, rows) {
        if (err) {
            throw logger.error('accesscontrol.GetAccessControlUsers: SELECT users - Error (Read)=' + err);
        }
        
        // Repopulate AccessControlUsers array        
        AccessControlUsers = [];
        for (i = 0; i < rows.length; i++) {
            AccessControlUsers.push({
                "Id"            : rows[i].Id,
                "Name"          : rows[i].Name,
                "Enabled"       : rows[i].Enabled,
                "Entries"       : rows[i].Entries,
                "Code"          : rows[i].Code,
                "TimeChanged"   : rows[i].TimeChanged
            });
        }
        logger.trace('accesscontrol.GetAccessControlUsers: Users=' + JSON.stringify(AccessControlUsers));
        
        callback(AccessControlUsers);
    });
};

// Get Access Control InvalidEntries
exports.GetAccessControlInvalidEntries = function (io, logger, callback) {
    // Get the number of InvalidEntries
    logger.trace('accesscontrol.GetAccessControlUsers: InvalidEntries=' + InvalidEntries);
    callback({ "InvalidEntries": InvalidEntries });
};

// Add an Access Control User
exports.AddAccessControlUser = function (req, db, io, logger, callback) {
    
    // Add Access Control User
    var TimeChanged = moment();
    var dbcommand = "'" + req.body.Name + "', '" + req.body.Enabled + "', " + 0 + ", '" + req.body.Code + "', '" + TimeChanged + "'";
    logger.trace("accesscontrol.AddAccessControlUser: Modify=" + dbcommand);
    //accesslogger.info("User Added: User=" + req.body.Name);
    db.run("INSERT INTO AccessControlUsers (Name, Enabled, Entries, Code, TimeChanged) VALUES (" + dbcommand + ")", function (err) {
        if (err !== null) {
            throw logger.error('accesscontrol.AddAccessControlUser: User=' + req.params.user + ' - Error (Update)=' + JSON.stringify(err));
        }
        
        _this.GetAccessControlUsers(db, io, logger, function (AccessControlUsers) {
            callback(AccessControlUsers)
        });
    });
};

// Delete an existing Access Control User
exports.DeleteAccessControlUser = function (req, db, io, logger, callback) {
    
    // Delete Access Control User
    var dbcommand = "DELETE FROM AccessControlUsers WHERE rowid='" + req.body.Id + "'";
    logger.trace("accesscontrol.DeleteAccessControlUser: Modify=" + dbcommand);
    //accesslogger.info("User Deleted: User=" + req.body.Name);
    db.run(dbcommand, function (err) {
        if (err !== null) {
            throw logger.error('accesscontrol.DeleteAccessControlUser: User=' + req.params.user + ' - Error (Update)=' + JSON.stringify(err));
        }
        
        _this.GetAccessControlUsers(db, io, logger, function (AccessControlUsers) {
            callback(AccessControlUsers)
        });
    });

};

// Modify an existing Access Control User
exports.ModifyAccessControlUser = function (req, db, io, logger, callback) {
    
    // Modify Access Control User
    var TimeChanged = moment();
    var dbcommand = "UPDATE AccessControlUsers SET Name='" + req.body.Name + "', Enabled='" + req.body.Enabled + "', Code='" + req.body.Code + "', TimeChanged='" + TimeChanged + "' WHERE rowid='" + req.params.user + "'";
    logger.trace("accesscontrol.ModifyAccessControlUser: Modify=" + dbcommand);
    //accesslogger.info("User Modified: User=" + req.body.Name);
    db.run(dbcommand, function (err) {
        if (err !== null) {
            throw logger.error('accesscontrol.ModifyAccessControlUser: User=' + req.params.user + ' - Error (Update)=' + JSON.stringify(err));
        }
        
        _this.GetAccessControlUsers(db, io, logger, function (AccessControlUsers) {
            callback(AccessControlUsers)
        });
    });
};
//#endregion Access Control Users
