// Define the configuraton for the I/O
var ANALOG_INPUTS_LOCATION = '/sys/class/hwmon';
var ANALOG_INPUTS_CHIPS_PER_BOARD = 1;// Default = 2
var ANALOG_INPUTS_POINTS_PER_CHIP = 8;// Default = 16
var ANALOG_INPUTS_MAXIMUM_RANGE = 25;// Default = 25
var ANALOG_INPUTS_DEFAULT_LOW_THRESHOLD = 0;// Default = 0
var ANALOG_INPUTS_DEFAULT_HIGH_THRESHOLD = ANALOG_INPUTS_MAXIMUM_RANGE;// Default = ANALOG_INPUTS_MAXIMUM_RANGE
var ANALOG_INPUTS_RAWVALUE_OFFSET = -2.1796;// Default = -2.1796
var ANALOG_INPUTS_RAWVALUE_SLOPE = 0.0112;// Default = 0.0112
var ANALOG_INPUTS_CHANGE_FILTER = 3;// Default = 3 - this is the amount of change in the raw value we look at before we send socket update
var ANALOG_INPUTS_FIRST_READ = true;// Track the first read

//var TotalDuration = 0;
//var TotalReads = 0;

// These constants are for use as 
var ANALOG_INPUTS_IN_RANGE = 0;
var ANALOG_INPUTS_BELOW_THRESHOLD = 1;
var ANALOG_INPUTS_ABOVE_THRESHOLD = 2;

// Timing analysis    
var totalCycleTime = 0.0;
var averageCycleTime = 0.0;
var totalCycles = 0;
var analogInputsReadStart;
var AnalogInputPointsRead = 0;

// The following array holds, in memory, the current value (in raw bit values - 0 to 2500)
var analogInputsArray = [
    { "Address" : "0", "Name" : "Name", "RawValue": 0, "RangedValue": 0 , "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "1", "Name" : "Name", "RawValue": 0, "RangedValue": 0 , "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "2", "Name" : "Name", "RawValue": 0, "RangedValue": 0 , "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "3", "Name" : "Name", "RawValue": 0, "RangedValue": 0 , "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "4", "Name" : "Name", "RawValue": 0, "RangedValue": 0 , "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "5", "Name" : "Name", "RawValue": 0, "RangedValue": 0 , "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "6", "Name" : "Name", "RawValue": 0, "RangedValue": 0 , "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "7", "Name" : "Name", "RawValue": 0, "RangedValue": 0 , "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
];

// Requires
var fs = require('fs');
var moment = require('moment');

// Initialize the analog inputs upon system start
//      - Checks if database table exists, if not creates the table and fills it with default data for each point
//      - Else reads from the database and updates analogInputsArray
exports.InitializeAnalogInputs = function (db, io, logger, callback) {
    // Try getting the AnalogInputs table
    db.get("SELECT name FROM sqlite_master WHERE type='table' AND name='AnalogInputs'", function (err, rows) {
        if (err !== null) {
            logger.error('Analog Inputs - Initialize Error (Check)=' + JSON.stringify(err));
        }
        else if (rows === undefined) {
            // We read nothing, so let's create the table ...
            logger.info("Analog Inputs - Initialize by creating database table ...");
            db.exec("CREATE TABLE if not exists AnalogInputs (Address INTEGER, Name TEXT, RawValue INTEGER, RangedValue TEXT, ThresholdState INTEGER, Enabled TEXT, SendEvents TEXT, Range TEXT, LowThreshold TEXT, HighThreshold TEXT, Offset TEXT, HoldTime INTEGER, Qualifier INTEGER, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    logger.error('Analog Inputs - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    var timeChanged = moment();// Record the create time as the last time the I/O point value was changed
                    
                    // Created the table, now add data ...
                    db.run("BEGIN")
                    for (var i = 0; i < (ANALOG_INPUTS_CHIPS_PER_BOARD * ANALOG_INPUTS_POINTS_PER_CHIP); i++) {
                        // Initialize the array
                        UpdateAnalogInputConfiguration(i, "Analog Input " + (i + 1), 0, 0, 0, "TRUE", "TRUE", ANALOG_INPUTS_MAXIMUM_RANGE, ANALOG_INPUTS_DEFAULT_LOW_THRESHOLD, ANALOG_INPUTS_DEFAULT_HIGH_THRESHOLD, 0, 0, 0, timeChanged, io, logger);
                        analogInputsArray[i].TempValue = 0;
                        analogInputsArray[i].HoldTimerActive = false;
                        analogInputsArray[i].HoldTimer = timeChanged;
                        
                        // Initialize the database
                        insertStatement = i + ", 'Analog Input " + (i + 1) + "', 0, '0', 0, 'TRUE', 'TRUE', '" + ANALOG_INPUTS_MAXIMUM_RANGE + "', '" + ANALOG_INPUTS_DEFAULT_LOW_THRESHOLD + "', '" + ANALOG_INPUTS_DEFAULT_HIGH_THRESHOLD + "', '0', 0, 0, '" + timeChanged + "'";
                        //logger.trace('INSERT INTO AnalogInputs (Address, Name, RawValue, RangedValue, Enabled, SendEvents, Range, LowThreshold, HighThreshold, Offset, HoldTime, Qualifier, TimeChanged) VALUES (' + insertStatement + ');');
                        db.run('INSERT INTO AnalogInputs (Address, Name, RawValue, RangedValue, ThresholdState, Enabled, SendEvents, Range, LowThreshold, HighThreshold, Offset, HoldTime, Qualifier, TimeChanged) VALUES (' + insertStatement + ');', function (err) {
                            if (err !== null) {
                                logger.error('Analog Inputs - Initialize Error (Insert)=' + JSON.stringify(err));
                            }
                        });
                    }
                    db.run("COMMIT");
                    
                    // Analog Inputs table created
                    if (typeof callback == "function") {
                        logger.info('Analog Inputs - Initialized new database table');
                        callback();
                    }
                }
            });
        }
        else {
            // Write out current AnalogInputs table
            logger.info("Analog Inputs - Initialize by reading database ...");
            db.each("SELECT rowid as id, Address, Name, RawValue, RangedValue, ThresholdState, Enabled, SendEvents, Range, LowThreshold, HighThreshold, Offset, HoldTime, Qualifier, TimeChanged FROM AnalogInputs ORDER BY Address", function (err, row) {
                if (err !== null) {
                    logger.error('Analog Inputs - Initialize Error (Select)=' + JSON.stringify(err));
                }
                else {
                    UpdateAnalogInputConfiguration(row.Address, row.Name, row.RawValue, row.RangedValue, row.ThresholdState, row.Enabled, row.SendEvents, row.Range, row.LowThreshold, row.HighThreshold, row.Offset, row.HoldTime, row.Qualifier, row.TimeChanged, io, logger)
                    logger.trace("Analog Inputs - [Value INITIAL from dB] Analog Input[" + row.Address + "] '" + row.Name + "' RawValue=" + row.RawValue);
                }
            }, function () { 
                // Analog Inputs read from table
                if (typeof callback == "function") {
                    logger.info('Analog Inputs - Initialized from existing database table');
                    callback();
                }
            });           
        }
    });
},

// Adjust the analogInputsArray
//  This is only called during initialization of the dB and if the client makes a change
module.exports.UpdateAnalogInputConfiguration = UpdateAnalogInputConfiguration;
function UpdateAnalogInputConfiguration(Address, Name, RawValue, RangedValue, ThresholdState, Enabled, SendEvents, Range, LowThreshold, HighThreshold, Offset, HoldTime, Qualifier, TimeChanged, io, logger) {
    
    analogInputsArray[Address].Name = Name;
    analogInputsArray[Address].RawValue = RawValue;
    analogInputsArray[Address].RangedValue = RangedValue;
    analogInputsArray[Address].ThresholdState = ThresholdState;
    analogInputsArray[Address].Enabled = Enabled;
    analogInputsArray[Address].SendEvents = SendEvents;
    analogInputsArray[Address].Range = Range;
    analogInputsArray[Address].LowThreshold = LowThreshold;
    analogInputsArray[Address].HighThreshold = HighThreshold;
    analogInputsArray[Address].Offset = Offset;
    analogInputsArray[Address].HoldTime = HoldTime;
    analogInputsArray[Address].Qualifier = Qualifier;
    analogInputsArray[Address].QualifierValue = 0;
    analogInputsArray[Address].TempValue = RawValue;// Set to last known RawValue, will be fixed by first read
    analogInputsArray[Address].TimeChanged = TimeChanged;
    
    logger.info("UpdateAnalogInputConfiguration - analogInputsArray[" + Address + "]=" + JSON.stringify(analogInputsArray[Address]));
    
    sendValueAnalogInputs(io, logger);// Update the Analog Inputs configuration
}

function ConvertRawValueToValue(rawValue, logger) {
    value = Number(parseFloat(rawValue) * ANALOG_INPUTS_RAWVALUE_SLOPE + ANALOG_INPUTS_RAWVALUE_OFFSET);
    return value;
}

function ConvertValueToRawValue(value, logger) {
    rawValue = Math.round((parseFloat(value) - ANALOG_INPUTS_RAWVALUE_OFFSET) / ANALOG_INPUTS_RAWVALUE_SLOPE);
    //logger.debug("ConvertValueToRawVaule: value=" + value + "  rawValue=" + rawValue);
    return rawValue;
}

// Intervalled processing of analog inputs
var processCounter = 0;
var processTotalTime = 0;
exports.ProcessAnalogInputs = function (db, logger, io, netSnmp, ledManager) {
    timeStart = moment();
    
    //lastAnalogInputRead = (moment() - analogInputsReadStart) / 1000.0;
    //analogInputsReadStart = moment()
    var count = 0;
    var index = 0;
    AnalogInputPointsRead = 0;
    for (chipCounter = 0; chipCounter < ANALOG_INPUTS_CHIPS_PER_BOARD ; chipCounter++) {
        //chipStartPoint = Date.now();
        //logger.info("Process Analog Inputs - Processing Chip=" + chipCounter);
        for (inputCounter = 0; inputCounter < ANALOG_INPUTS_POINTS_PER_CHIP; inputCounter++) {
            //logger.debug('ProcessAnalogInputs: chip=' + chipCounter + ' input=' + inputCounter);
            readAnalogInputFile(chipCounter, inputCounter, db, io, logger, netSnmp)
        }
    }
    ANALOG_INPUTS_FIRST_READ = false;

    //if (processCounter === 500) {
    //    logger.info("ProcessAnalogInputs: Average process time = " + (processTotalTime / processCounter).toFixed(2) + "msec");
    //    processCounter = 0;
    //    processTotalTime = 0;
    //}
    //processCounter++;
    //processTotalTime += (Date.now() - timeStart);
}

function readAnalogInputFile(chipCounter, inputCounter, db, io, logger, netSnmp) {
    //logger.info("Reading: " + ANALOG_INPUTS_LOCATION + "/hwmon" + chipCounter + "/in" + inputCounter + "_input");
    
    index = inputCounter + ANALOG_INPUTS_POINTS_PER_CHIP * chipCounter;
    timeChanged = moment();
    
    // Only read value if the I/O point is enabled
    if (analogInputsArray[index].Enabled === "TRUE") {
        try {
            //timeStart = moment();
            var readValue = 0;
            try {
                data = fs.readFileSync(ANALOG_INPUTS_LOCATION + "/hwmon" + chipCounter + "/in" + inputCounter + "_input");
                readValue = String(data).replace(/(\r\n|\n|\r|[ +])/gm, "");
                //logger.debug("Reading " + ANALOG_INPUTS_LOCATION + "/hwmon" + chipCounter + "/in" + inputCounter + "_input - readValue=" + readValue);
                //logger.debug('Analog Inputs - Process OK (File Read)= Normal (Analog Input at Address=' + index + ' readValue error=' + readValue + ')');
            }
            // This is to catch infrequent errors related to reading
            catch (error) {
                // ***************** We need to address this issue big time ******************
                logger.error('Analog Inputs - Process Error (File Read)=' + error + ' (Analog Input at Address=' + index + ' chip=' + chipCounter + ' input=' + inputCounter + ')');
                return;
            }
            
            //logger.info("readAnalogInputFile: Index=" + index + " readValue=" + readValue + " oldValue=" + analogInputsArray[index].RawValue);
            //analogInputReadEnd = Date.now() - analogInputsReadStart;
            //Duration = (Date.now() - timeStart);
            //TotalDuration += Duration;
            //TotalReads++;
            //AverageDuration = TotalDuration / TotalReads;
            //logger.info("Finished Reading: " + ANALOG_INPUTS_LOCATION + "/hwmon" + chipCounter + "/in" + inputCounter + "_input - Value=" + readValue + ", Duration=" + Duration + "ms" + ", Average Duration=" + AverageDuration.toFixed(2)+ "ms");
            
            // Add the offset to the read value
            readValue = Math.round(Number(readValue) + Number(ConvertValueToRawValue(analogInputsArray[index].Offset, logger)) - Number(ConvertValueToRawValue(0, logger)));
            
            // Create temp value to track allow us to handle hold times & qualifiers
            tempAnalogInputPoint = JSON.parse(JSON.stringify(analogInputsArray[index]));
            tempAnalogInputPoint.RawValue = readValue;
            tempAnalogInputPoint.RangedValue = ConvertRawValueToValue(readValue, logger);            

            // Send even small changes to clients regardless of anything of hold times or qualifiers
            if (Math.abs(analogInputsArray[index].TempValue - readValue) > ANALOG_INPUTS_CHANGE_FILTER) {
                tempAnalogInputPoint.TimeChanged = timeChanged;
                sendValueAnalogInput("AnalogInput", tempAnalogInputPoint, io, logger);
                rangedValue = ConvertRawValueToValue(tempAnalogInputPoint.RawValue, logger).toFixed(2);
                logger.info("Analog Inputs - [Value CHANGED] Analog Input[" + tempAnalogInputPoint.Address + "] '" + tempAnalogInputPoint.Name + "' RawValue=" + tempAnalogInputPoint.RawValue + " TempValue=" + tempAnalogInputPoint.TempValue + " ConvertedValue=" + rangedValue + " ThresholdState=" + tempAnalogInputPoint.ThresholdState );
            }
            analogInputsArray[index].TempValue = readValue;// Remember the last value, for the next rawValue read
            
            // Check to see if we have crossed a threshold boundary
            //  Compare last good stored RawValue to this read's RawValue to see if threshold has changed
            //  And also check to see if the ThresholdState is < 0 (which means that we need to force a re-evaluation of the ThresholdState due to configuration changes
            if (getAnalogInputThresholdState(analogInputsArray[index], analogInputsArray[index].RawValue, logger) != getAnalogInputThresholdState(tempAnalogInputPoint, tempAnalogInputPoint.RawValue, logger)
                || (analogInputsArray[index].ThresholdState < 0)) {
                oldThreshold = analogInputsArray[index].ThresholdState;
                newThreshold = getAnalogInputThresholdState(tempAnalogInputPoint, tempAnalogInputPoint.RawValue, logger);
                
                //  Check Qualifier if there is one
                if (analogInputsArray[index].Qualifier > 0) {
                    // logger.info("   Qualifier Exists for Analog Input[" + tempAnalogInputPoint.Address + "]  using Digital Input " + (analogInputsArray[index].Qualifier - 1) + " which currently has value=" + analogInputsArray[index].QualifierValue + ")");
                    if (analogInputsArray[index].QualifierValue > 0) {
                        // The change *IS* qualified, so proceed
                        if (analogInputsArray[index].HoldTimerActive === false) {
                            //  All things go, we can now start the hold timer
                            analogInputsArray[index].HoldTimer = timeChanged;                        
                            analogInputsArray[index].HoldTimerActive = true;
                            logger.info("Analog Inputs - [Threshold CHANGE DETECTED, Hold Time started] Analog Input[" + tempAnalogInputPoint.Address + "] using qualifier DigitalInput[" + (analogInputsArray[index].Qualifier + 1) + "]" + " - Previous Threshold="+oldThreshold+" New Threshold="+newThreshold);
                        }
                    } else {
                        //  The change is *NOT* qualified, so ignore
                        if (analogInputsArray[index].HoldTimerActive === true && ANALOG_INPUTS_FIRST_READ === false)
                            logger.info("Analog Inputs - [Threshold CHANGE NEGATED due to qualifier failure] Analog Input[" + tempAnalogInputPoint.Address + "] using qualifier DigitalInput[" + (analogInputsArray[index].Qualifier + 1) + "]" + " - Previous Threshold=" + oldThreshold + " New Threshold=" + newThreshold);
                        analogInputsArray[index].HoldTimerActive = false;
                    }
                } else {
                    // Start the hold timer for inputs that have no qualifier
                    analogInputsArray[index].HoldTimer = timeChanged;
                    analogInputsArray[index].HoldTimerActive = true;
                    logger.info("Analog Inputs - [Threshold CHANGE DETECTED, Hold Time Started] Analog Input[" + tempAnalogInputPoint.Address + "] using *NO* Qualifier" + " - Previous Threshold=" + oldThreshold + " New Threshold=" + newThreshold);
                }
            } else {
                // There was no threshold change, so just ignore this as a input change for now
                oldThreshold = analogInputsArray[index].ThresholdState;
                if (analogInputsArray[index].HoldTimerActive === true && ANALOG_INPUTS_FIRST_READ===false)
                    logger.info("Analog Inputs - [Threshold CHANGE NEGATED due to back in range] Analog Input[" + tempAnalogInputPoint.Address + "] using qualifier DigitalInput[" + (analogInputsArray[index].Qualifier - 1) + "]" + " - Previous Threshold=" + oldThreshold);
                analogInputsArray[index].HoldTimerActive = false;
            }
            
            //logger.info("ProcessAnalogInputs: Assess Change @ Index=" + index + " TempValue='" + analogInputsArray[index].TempValue + "' RawValue='" + analogInputsArray[index].RawValue + "'");
            //logger.info("ProcessAnalogInputs: Assess Change @ Index=" + index + " HoldTimer='" + analogInputsArray[index].HoldTimer + "' timeChanged='" + timeChanged + "'")
            //logger.info("ProcessAnalogInputs: Assess Change @ Index=" + index + " Timer difference='" + (timeChanged - analogInputsArray[index].HoldTimer) / 1000 + "' HoldTime='" + analogInputsArray[index].HoldTime + "'");
            
            // Record changes only if:
            //      HoldTimer is OK
            //      Qualifier is OK (note: HoldTimerActive value of true also indicates that qualifier passed)
            if ((analogInputsArray[index].HoldTimerActive == true) && ((timeChanged - analogInputsArray[index].HoldTimer) / 1000 >= analogInputsArray[index].HoldTime)) {
                // Satisfied the hold time, so update the actual value
                //logger.info("ProcessAnalogInputs: Record Change @ Index=" + index + " Updated Value='" + analogInputsArray[index].TempValue + "'  Old Value='" + analogInputsArray[index].TempValue + "'");
                analogInputsArray[index].RawValue = readValue;// Capture the most current value
                analogInputsArray[index].TimeChanged = timeChanged;
                analogInputsArray[index].ThresholdState = getAnalogInputThresholdState(analogInputsArray[index], analogInputsArray[index].RawValue, logger)
                analogInputsArray[index].RangedValue = ConvertRawValueToValue(analogInputsArray[index].RawValue, logger).toFixed(2);
                analogInputsArray[index].HoldTimerActive = false;
                sendSocketAndSnmp(analogInputsArray[index], db, io, logger, netSnmp);
                //logger.info("Analog Inputs - [Value HELD and QUALIFIED] Analog Input[" + tempAnalogInputPoint.Address + "] '" + tempAnalogInputPoint.Name + "' RawValue=" + tempAnalogInputPoint.RawValue + " TempValue=" + readValue + " ConvertedValue=" + tempAnalogInputPoint.RangedValue.toFixed(2) + " Hold Timer Active=" + analogInputsArray[index].HoldTimerActive);
                
                // Update the database and then when done, send changes to socket and SNMP clients
                db.run("UPDATE AnalogInputs SET RawValue='" + analogInputsArray[index].RawValue 
                    + "', TimeChanged='" + analogInputsArray[index].TimeChanged 
                    + "', RangedValue='" + analogInputsArray[index].RangedValue 
                    + "', ThresholdState='" + analogInputsArray[index].ThresholdState 
                    + "' WHERE Address = '" + index + "'", 
                    function () { });
            }
                //logger.info("AnalogInput[" + index + "]=" + readValue + " (Enabled)"
        }
        catch (error) {
            logger.error('Analog Inputs - Process Error (General)=' + error);
            logger.error('Analog Inputs - Address=' + index + '  previousValue=' + analogInputsArray[index].RawValue + ' readValue=' + readValue);
        }
    }
    else {
    //logger.info("AnalogInput[" + index + "]= " + analogInputsArray[index].RawValue + " (Disabled)")
    }
}

// Use this function to force a read of the I/O
//  We need to do this after we change a configuration
//  We force both a read of the RawValue and ThresholdState
//  This function basically forces a new evaulation, as a result socket and SNMP messages
module.exports.ForceNewRawValue = ForceNewRawValue;
function ForceNewRawValue(index, logger) {
    logger.trace("ForceNewRawValue: For address index=" + index);
    analogInputsArray[index].TempValue = -1;
    analogInputsArray[index].ThresholdState = -1;
}

module.exports.UpdateQualifierValues = UpdateQualifierValues;
function UpdateQualifierValues(digitalInputsArray, db, logger, io) {
    for (qualifier = 0; qualifier < digitalInputsArray.length; qualifier++) {
        for (analogInput = 0; analogInput < analogInputsArray.length ; analogInput++) {
            if (analogInputsArray[analogInput].Qualifier > 0) {
                // This analog has a qualifier defined, so let's update the current qualifier value
                if ((analogInputsArray[analogInput].Qualifier - 1) == qualifier) {
                    if (analogInputsArray[analogInput].QualifierValue != digitalInputsArray[qualifier].Value)
                        logger.info("    Analog Input[" + analogInput + "] qualifier Digital["+(qualifier-1)+ "] has changed value to " + digitalInputsArray[qualifier].Value);
                    analogInputsArray[analogInput].QualifierValue = digitalInputsArray[qualifier].Value;
                }
            }
        }
    }
}

module.exports.getAnalogInputCurrentValue = getAnalogInputCurrentValue;
function getAnalogInputCurrentValue(address, logger) {
    logger.trace("getAnalogInputCurrentValue - address=" + address + " value=" + analogInputsArray[address].TempValue);
    return analogInputsArray[address].TempValue;
}

module.exports.getAnalogInputThresholdState = getAnalogInputThresholdState;
function getAnalogInputThresholdState(AnalogInput, Value, logger) {
    lowThresholdRawValue = ConvertValueToRawValue(AnalogInput.LowThreshold, logger);
    highThresholdRawValue = ConvertValueToRawValue(AnalogInput.HighThreshold, logger);
    
    tempValueState = ANALOG_INPUTS_IN_RANGE;
    if (Value < lowThresholdRawValue) {
        tempValueState = ANALOG_INPUTS_BELOW_THRESHOLD
    } else if (Value > highThresholdRawValue) {
        tempValueState = ANALOG_INPUTS_ABOVE_THRESHOLD
    }
    
    return tempValueState;
}

// Initialize the snmpHostSessionsArray used for send SNMP messages
var SnmpHostSessionsArray = [];// Instance array for list of SNMP host sessions
exports.SetSnmpHostSessionsArray = function (snmpHostSessions, logger, callback) {
    SnmpHostSessionsArray = snmpHostSessions.slice(0);// Copy array
    logger.trace("Analog Inputs - Initialized SNMP Host Sessions");
    
    if (typeof callback == "function") {
        callback();
    }
}

function sendSocketAndSnmp(analogInput, db, io, logger, netSnmp) {
    if (analogInput.Qualifier > 0) {
        if (analogInput.HoldTime > 0) {
            logger.info("Analog Inputs - [Threshold CHANGED COMPLETED with HoldTime & Qualifier] Analog Input[" + analogInput.Address + "] '" + analogInput.Name + "' RawValue=" + analogInput.RawValue + " TempValue=" + analogInput.TempValue + " ConvertedValue=" + analogInput.RangedValue + " New Threshold State=" + analogInput.ThresholdState);
            logger.trace("Analog Inputs - [Threshold CHANGED COMPLETED with HoldTime & Qualifier] Analog Input[" + analogInput.Address + "] =" + JSON.stringify(analogInput));
        } else {
            logger.info("Analog Inputs - [Threshold CHANGED COMPLETED with Qualifier] Analog Input[" + analogInput.Address + "] '" + analogInput.Name + "' RawValue=" + analogInput.RawValue + " TempValue=" + analogInput.TempValue + " ConvertedValue=" + analogInput.RangedValue  + " New Threshold State=" + analogInput.ThresholdState);
            logger.trace("Analog Inputs - [Threshold CHANGED COMPLETED with Qualifier] Analog Input[" + analogInput.Address + "] =" + JSON.stringify(analogInput));
        }
    } else {
        if (analogInput.HoldTime > 0) {
            logger.info("Analog Inputs - [Threshold CHANGED COMPLETED with HoldTime] Analog Input[" + analogInput.Address + "] '" + analogInput.Name + "' RawValue=" + analogInput.RawValue + " TempValue=" + analogInput.TempValue + " ConvertedValue=" + analogInput.RangedValue  + " New Threshold State=" + analogInput.ThresholdState);
            logger.trace("Analog Inputs - [Threshold CHANGED COMPLETED with HoldTime] Analog Input[" + analogInput.Address + "] =" + JSON.stringify(analogInput));
        } else {
            logger.info("Analog Inputs - [Threshold CHANGED COMPLETED] Analog Input[" + analogInput.Address + "] '" + analogInput.Name + "' RawValue=" + analogInput.RawValue + " TempValue=" + analogInput.TempValue + " ConvertedValue=" + analogInput.RangedValue  + " New Threshold State=" + analogInput.ThresholdState);
            logger.trace("Analog Inputs - [Threshold CHANGED COMPLETED] Analog Input[" + analogInput.Address + "] =" + JSON.stringify(analogInput));
        }    
    }
    
    // SNMP: Send SNMP Trap/Notification if threshold(s) crossed
    sendSNMP(analogInput, io, logger, netSnmp)
    
    // SOCKET.IO: Send changed value to all attached clients
    sendValueAnalogInput("AnalogInput", analogInput, io, logger);
}

function sendValueAnalogInput(dataType, data, io, logger) {
    //logger.info("sendValueAnalogInput - [Value UPDATED] Analog Input[" + data.Address + "] '" + data.Name + "' RawValue=" + data.RawValue);
    io.sockets.emit(dataType, data);
}

function sendValueAnalogInputs(io, logger) {
    enabled = 0;
    for (i = 0; i < analogInputsArray.length; i++) {
        if (analogInputsArray[i].Enabled === "TRUE") enabled++;
    }
    io.sockets.emit("AnalogInputs", { "AnalogInputsTotal": analogInputsArray.length, "AnalogInputsEnabled": enabled })
}

function sendSNMP(ioPoint, io, logger, netSnmp) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (ioPoint.Enabled == "FALSE") return;
    if (ioPoint.SendEvents == "FALSE") return;
    
    address = parseInt(ioPoint.Address) + 1;// To overcome concatenation
    logger.info("Analog Inputs - Sending SNMP Trap for Analog Input[" + address + "]");

    // Set up the SNMP trap    
    var trapOid = "1.3.6.1.4.1.11902.3.3.1000.1000.0." + (address - 1 + 1000);
    var varbinds = [
        {
            // Timestamp
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.1",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(moment().toString())
        },
        {
            // Address
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.2",
            type: netSnmp.ObjectType.Integer,
            value: parseInt(ioPoint.Address) + 1
        },
        {
            // Name
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.3",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(ioPoint.Name)
        },
        {
            // RangedValue
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.4",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(((ConvertRawValueToValue(ioPoint.RawValue, logger)).toFixed(2)).toString())
        },
        {
            // ThresholdState
            //  ANALOG_INPUTS_IN_RANGE = 0
            //  ANALOG_INPUTS_BELOW_THRESHOLD = 1
            //  ANALOG_INPUTS_ABOVE_THRESHOLD = 2
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.5",
            type: netSnmp.ObjectType.Integer,
            value: parseInt(ioPoint.ThresholdState)
        },
        {
            // LowThreshold
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.6",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(ioPoint.LowThreshold)
        },
        {
            // HighThreshold
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.7",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(ioPoint.HighThreshold)
        },
        {
            // Hold Time
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.8",
            type: netSnmp.ObjectType.Integer,
            value: parseInt(ioPoint.HoldTime)
        },
        {
            // Offset (in 100mV)
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.9",
            type: netSnmp.ObjectType.Integer,
            value: parseInt(ioPoint.Offset * 100)
        },
        {
            // Qualifier
            oid: "1.3.6.1.4.1.11902.3.3." + (address) + ".0.10",
            type: netSnmp.ObjectType.Integer,
            value: parseInt(ioPoint.Qualifier)
        }
    ];
    
    // Send the SNMP notification
    for (i = 0; i < SnmpHostSessionsArray.length; i++) {
        // Skip if this host is not enabled
        if (SnmpHostSessionsArray[i].Enabled === "FALSE") continue;
        
        // Send trap to host ...
        IPAddress = SnmpHostSessionsArray[i].Session.target;
        SnmpHostSessionsArray[i].Session.trap(trapOid, varbinds, function (error) {
            if (error)
                logger.error("Analog Inputs - Send SNMP Error=" + error);
            else
                logger.trace("Analog Inputs - Sending to SNMP Host @ " + IPAddress + " - Analog Input[" + varbinds[1].value + "] '" + varbinds[2].value + "' Value=" + varbinds[4].value);
        });
    }
}

exports.GetRawValue = function (point) {
    return analogInputsArray[point].TempValue;
}