// Define the configuraton for the I/O
var ANALOG_OUTPUTS_LOCATION = '/sys/class/hwmon/hwmon2';
var ANALOG_OUTPUTS_POINTS = 4;//default = 4
var ANALOG_OUTPUTS_DEFAULT_RANGE = 20;// default = 20
var ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE = 200;
var ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET = 0;

// The following array holds, in memory, the current value (in raw bit values - 0 to 2048)
var analogOutputsArray = [
    { "Address" : "0", "Name" : "Name",  "RawValue": 0, "RangedValue": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": 20, "TimeChanged": 0 }, 
    { "Address" : "1", "Name" : "Name",  "RawValue": 0, "RangedValue": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": 20, "TimeChanged": 0 }, 
    { "Address" : "2", "Name" : "Name",  "RawValue": 0, "RangedValue": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": 20, "TimeChanged": 0 }, 
    { "Address" : "3", "Name" : "Name",  "RawValue": 0, "RangedValue": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": 20, "TimeChanged": 0 }
];

// Requires for the I/O
var fs = require('fs');
var moment = require('moment');

// Initialize the analog outputs upon system start
//      - Checks if database table exists, if not creates the table and fills it with default data for each point
//      - Else reads from the database and updates analogOutputsArray
exports.InitializeAnalogOutputs = function (db, io, logger, callback) {
    // Try getting the AnalogOutputs table
    db.get("SELECT name FROM sqlite_master WHERE type='table' AND name='AnalogOutputs'", function (err, rows) {
        if (err !== null) {
            logger.error('Analog Outputs - Initialize Error (Check)=' + JSON.stringify(err));
        }
        else if (rows === undefined) {
            // We read nothing, so let's create the table ...
            logger.info("Analog Outputs - Initialize by creating database table ...");
            db.exec("CREATE TABLE if not exists AnalogOutputs (Address INTEGER, Name TEXT, RawValue INTEGER, RangedValue TEXT, Enabled TEXT, SendEvents TEXT, Range TEXT, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    logger.error('Analog Outputs - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    var timeChanged = moment();// Record the create time as the last time the I/O point value was changed
                    
                    // Created the table, now add data ...
                    db.run("BEGIN")
                    for (var i = 0; i < ANALOG_OUTPUTS_POINTS; i++) {
                        // Initialize the array
                        UpdateAnalogOutputConfiguration(i, "Analog Ouput " + (i+1), 0, 0, "TRUE", "TRUE", ANALOG_OUTPUTS_DEFAULT_RANGE, timeChanged, io, logger);
                        // Empty space here to be consistent
                        // Empty space here to be consistent
                        
                        // Initialize the database
                        insertStatement = i + ", 'Analog Output " + (i+1) + "', 0, '0', 'TRUE', 'TRUE', '" + ANALOG_OUTPUTS_DEFAULT_RANGE + "', '" + timeChanged + "'";
                        db.run("INSERT INTO AnalogOutputs (Address, Name, RawValue, RangedValue, Enabled, SendEvents, Range, TimeChanged) VALUES (" + insertStatement + ")", function (err) {
                            if (err !== null) {
                                logger.error('Analog Outputs - Initialize Error (Insert)=' + JSON.stringify(err));
                            }
                        });
                    }
                    db.run("COMMIT");

                    // Analog Outputs table created
                    if (typeof callback == "function") {
                        logger.info('Analog Outputs - Initialized new database table');
                        callback();
                    }
                }
            });
        }
        else {
            logger.info("Analog Outputs - Initialize by reading database ...");
            // Write out current AnalogOutputs table
            db.each("SELECT rowid as id, Address, Name, RawValue, RangedValue, Enabled, SendEvents, Range, TimeChanged FROM AnalogOutputs ORDER BY Address", function (err, row) {
                if (err !== null) {
                    logger.error('Analog Outputs - Initialize Error (Select)=' + JSON.stringify(err));
                }
                else {
                    UpdateAnalogOutputConfiguration(row.Address, row.Name, row.RawValue, row.RangedValue, row.Enabled, row.SendEvents, row.Range, row.TimeChanged, io, logger)
                }
            }, function () { 
                // Analog Outputs read from table
                if (typeof callback == "function") {
                    logger.info('Analog Outputs - Initialized from existing database table');
                    callback();
                }
            });
        }
    });
};

// Adjust the analogOutputsArray
module.exports.UpdateAnalogOutputConfiguration = UpdateAnalogOutputConfiguration;
function UpdateAnalogOutputConfiguration(Address, Name, RawValue, RangedValue, Enabled, SendEvents, Range, TimeChanged, io, logger) {
    analogOutputsArray[Address].Name = Name;
    analogOutputsArray[Address].RawValue = RawValue;
    analogOutputsArray[Address].RangedValue = RangedValue;
    analogOutputsArray[Address].Enabled = Enabled;
    analogOutputsArray[Address].SendEvents = SendEvents;
    analogOutputsArray[Address].Range = Range;
    analogOutputsArray[Address].TimeChanged = TimeChanged;
    sendValueAnalogOutputs(io, logger);// Update the Analog Outputs configuration
}

exports.ResetAnalogOutputs = function (db, logger, io, netSnmp, ledManager, callback) {
    var timeChanged = moment();
    for (index = 0; index < ANALOG_OUTPUTS_POINTS; index++) {
        // Reset the database point ...
        // We may not have to do this database update
        //db.run("UPDATE AnalogOutputs SET RawValue='" + 0 + "', TimeChanged='" + timeChanged + "' WHERE Address = '" + index + "'", 
        //    sendSocketAndSnmp(analogOutputsArray[index], db, io, logger, netSnmp));
        
        // Reset the output point ...        
        fs.writeFile(ANALOG_OUTPUTS_LOCATION + "/out" + index + "_output", '0', function (err) {
            if (err) {
                return logger.error('Analog Outputs - Reset Error (Write)=' + JSON.stringify(err));
            }
        });
    }

    // Analog Outputs reset is complete
    if (typeof callback == "function") {
        logger.trace('Analog Outputs - Reset to initial values');
        callback();
    }
}

// Process the I/O
var processCounter = 0;
var processTotalTime = 0;
exports.ProcessAnalogOutputs = function (db, logger, io, netSnmp, ledManager) {
    //timeStart = moment();
    
    var index = 0;// This is an index to analog I/O point 
    var fileIndex = 0;// This is an index to the GPIO location
    var timeChanged = moment();
    for (index = 0; index < ANALOG_OUTPUTS_POINTS; index++) {
        fileIndex = index;
        
        // Only read value if the I/O point is enabled
        if (analogOutputsArray[index].Enabled === "TRUE") {
            try {
                data = fs.readFileSync(ANALOG_OUTPUTS_LOCATION + "/out" + fileIndex + "_output");
                readValue = String(data).replace(/(\r\n|\n|\r)/gm, "");
                
                // Check to see if value has changed, if so update database and clients ...
                if (analogOutputsArray[index].RawValue != readValue) {
                    analogOutputsArray[index].RawValue = readValue;
                    analogOutputsArray[index].TimeChanged = timeChanged;
                    sendSocketAndSnmp(analogOutputsArray[index], db, io, logger, netSnmp);

                    // Value has changed, update the database and then the clients ...
                    db.run("UPDATE AnalogOutputs SET RawValue='" + readValue 
                        + "', TimeChanged='" + timeChanged 
                        + "' WHERE Address = '" + index + "'", function () {
                    });
                }
                //logger.info("AnalogOutput[" + index + "]=" + readValue + " (Enabled)")
            }
			catch (error) {
                logger.error('Analog Outputs - Process Error (Read)=' + JSON.stringify(error));
            }
        }
        else {
            //logger.info("AnalogOutput[" + index + "]=" + analogOutputsArray[index].Value + " (Disabled)")
        }
    }
    //if (processCounter === 500) {
    //    logger.info("ProcessAnalogOutputs: Average process time = " + (processTotalTime / processCounter).toFixed(2) + "msec");
    //    processCounter = 0;
    //    processTotalTime = 0;
    //}
    //processCounter++;
    //processTotalTime += (Date.now() - timeStart);
}

// Initialize the snmpHostSessionsArray used for send SNMP messages
var SnmpHostSessionsArray = [];// Instance array for list of SNMP host sessions
exports.SetSnmpHostSessionsArray = function (snmpHostSessions, logger, callback) {
    SnmpHostSessionsArray = snmpHostSessions.slice(0);// Copy array
    logger.trace("Analog Outputs - Initialized SNMP Host Sessions");
    
    if (typeof callback == "function") {
        callback();
    }
}

function sendSocketAndSnmp(analogOutput, db, io, logger, netSnmp) {
    rangedValue = ((analogOutputsArray[analogOutput.Address].RawValue - ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET)/ ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE).toFixed(2)
    logger.info("Analog Outputs - [Value UPDATED] Analog Output[" + analogOutput.Address + "] '" + analogOutput.Name + "' Value=" + rangedValue);
    logger.trace("Analog Outputs - [Value UPDATED] Analog Output[" + analogOutput.Address + "] =" + JSON.stringify(analogOutput));

    // SOCKET.IO: Send changed value to all attached clients
    sendValueAnalogOutput("AnalogOutput", analogOutput, io, logger);
    
    // SNMP: Send SNMP Trap/Notification if threshold(s) crossed
    sendSNMP(analogOutput, io, logger,netSnmp);
}

function sendValueAnalogOutput(dataType, data, io, logger) {
    io.sockets.emit(dataType, data);
}

function sendValueAnalogOutputs(io, logger) {
    enabled = 0;
    for (i = 0; i < analogOutputsArray.length; i++) {
        if (analogOutputsArray[i].Enabled === "TRUE") enabled++;
    }
    io.sockets.emit("AnalogOutputs", { "AnalogOutputsTotal": analogOutputsArray.length, "AnalogOutputsEnabled": enabled })
}

function sendSNMP(ioPoint, io, logger, netSnmp) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (ioPoint.Enabled == "FALSE") return;
    if (ioPoint.SendEvents == "FALSE") return;
    
    address = parseInt(ioPoint.Address) + 1;// To overcome concatenation
    logger.info("Analog Outputs- Sending SNMP Trap for Analog Output[" + address + "]");
    
    // Set up the SNMP trap    
    var trapOid = "1.3.6.1.4.1.11902.3.4.1000.1000.0." + (address - 1 + 1000);
    var varbinds = [
        {
            // Timestamp
            oid: "1.3.6.1.4.1.11902.3.4." + (address) + ".0.1",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(moment().toString())
        },
        {
            // Address
            oid: "1.3.6.1.4.1.11902.3.4." + (address) + ".0.2",
            type: netSnmp.ObjectType.Integer,
            value: parseInt(ioPoint.Address)+1
        },
        {
            // Name
            oid: "1.3.6.1.4.1.11902.3.4." + (address) + ".0.3",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(ioPoint.Name)
        },
        {
            // RangedValue
            oid: "1.3.6.1.4.1.11902.3.4." + (address) + ".0.4",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(((ioPoint.RawValue - ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET)/ ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE).toFixed(2)).toString()
        }
    ];
    
    // Send the SNMP notification
    for (i = 0; i < SnmpHostSessionsArray.length; i++) {
        // Skip if this host is not enabled
        if (SnmpHostSessionsArray[i].Enabled === "FALSE") continue;
        
        // Send trap to host ...
        IPAddress = SnmpHostSessionsArray[i].Session.target;
        SnmpHostSessionsArray[i].Session.trap(trapOid, varbinds, function (error) {
            if (error)
                logger.error("Analog Outputs - Send SNMP Error=" + error);
            else
                logger.trace("Analog Outputs - Sending to SNMP Host @ " + IPAddress + " - Analog Output[" + varbinds[1].value + "] '" + varbinds[2].value + "'");
        });
    }
}